# -*- coding: utf-8 -*-
"""3D effective point polarizabilities

.. currentmodule:: torchgdm.struct.point

.. autosummary::
   :toctree: generated/
   
   StructEffPola3D
   StructMieSphereEffPola3D
"""
from .pola import StructEffPola3D
from .pola import StructMieSphereEffPola3D
from . import pola
