# -*- coding: utf-8 -*-
"""torchgdm postprocessing modules

.. currentmodule:: torchgdm.postproc

.. autosummary::
   :toctree: generated/

   crosssect
   fields
   multipole
   green

"""
from . import crosssect
from . import fields
from . import multipole
from . import green